/*! ********************************************
 * SynthesizerWT.hpp
 * Copyright 2019 masahiro takamoto (MotChan)
 * Distributed under the MIT license. 
 * see -> sora/license.txt
 ******************************************** */
#ifndef INCLUDE_SynthesizerWT_h
#define INCLUDE_SynthesizerWT_h

#ifndef INCLUDE_MultiTimberSynthBase_h
#include "MultiTimberSynthBase.hpp"
#endif

#ifndef INCLUDE_SampleData_h
#include "SampleData.hpp"
#endif


namespace sora {
namespace graybox {
namespace synth {
/***********************************************
	Class	TonePresetWT
************************************************/
class TonePresetWT
{
public:
	char toneName[17]={"                "};	// ASCII 16Chars 
	int32 key = 64;        // 0 - 127     *Percussion
	int32 pan = 64;        // 0 - 127     *Percussion
	int32 lfoFreq = 2048;  // 0 - 16383   LFO - Phase Update Timer Parameter
	int32 strc = 0;        // 0 - 3
	int32 tune = 8192;     // 0 - 16383   center 8192, octave = 8192
	int32 lms = 0;         // 0 - 15      LFO - Modulation Depth Senstivity
	int32 wave = 0;        // 0 - 127
	int32 tl = 0;          // 0 - 127
	int32 ar = 127;        // 0 - 127
	int32 dr = 100;        // 0 - 127
	int32 sl = 0;          // 0 - 127
	int32 rr = 120;        // 0 - 127
	int32 kf = 0;          // 0 - 3       Key Follow Sensitivity         *Instrument 
	int32 kfStart = 48;    // 0 - 127     Key Number - Start Key Follow  *Instrument
	
	void copy(const TonePresetWT& preset);
	void reset();
};
/***********************************************
	Class	SynthesizerWT
************************************************/
class SynthesizerWT final : public MultiTimberSynthBase
{
private:
	static constexpr int32 WAVET_CONT_SIZE_BITS = 7;
	static constexpr int32 WAVET_CONT_SIZE = 1 << WAVET_CONT_SIZE_BITS;
	static constexpr int32 PG_PHASE_FP = 18;  // 18bit fixed-point  (precision 1 / 262144)
/***********************************************
	Class	SynthesizerWT :: VoiceUnitWT
************************************************/
private:
	class VoiceUnitWT final : public VoiceUnitBase
	{
	// Tables
	private:
		// Log2 Volume Table
		float* _LOG2V_TABLE = nullptr;
		// Pow2 Table
		static constexpr int32 _POW2T_OCTV_BITS = SynthData::POW2T_OCTV_BITS;
		static constexpr int32 _POW2T_OCTV_MASK = SynthData::POW2T_OCTV_MASK;
		int32* _POW2_TABLE = nullptr;
		int32* _POW2T_RS_TABLE = nullptr;
		int32* _POW2T_KEY_TABLE = nullptr;
		// EG
		static constexpr int32 _EG_PRAM_OCTV_LS = SynthData::POW2T_OCTV_BITS - 3;
		static constexpr int32 _EG_CNT_FP = SynthData::EG_CNT_FP;
		static constexpr int32 _EG_MAX_CNT = SynthData::EG_MAX_CNT;
		static constexpr int32 _EG_CNT_ENV_ATK = SynthData::EG_CNT_ENV_ATK;
		static constexpr int32 _EG_CNT_ENV_BTM = SynthData::EG_CNT_ENV_BTM;
		static constexpr int32 _EG_SUSTAIN_BTM = SynthData::EG_SUSTAIN_BTM;
		int32* _EG_TOTAL_LEVEL_TABLE = nullptr;
		int32* _EG_LEVEL_TABLE_UP = nullptr;
		int32* _EG_LEVEL_TABLE_DW = nullptr;
		int32* _EG_KEY_TABLE = nullptr;
		int32* _EG_POW2T_RS_TABLE = nullptr;
		int32* _EG_CNT_GATE_TABLE = nullptr;
		// PG
		//static constexpr int32 _WAVET_CONT_SIZE_BITS = SynthesizerWT::WAVET_CONT_SIZE_BITS;
		//static constexpr int32 _PG_PHASE_FP      = SynthesizerWT::PG_PHASE_FP;
		static constexpr int32 _PG_PHASE_FP_MASK = (1 << PG_PHASE_FP) - 1;
		// LFO
		static constexpr int32 _LFO_TIMER_INIT_VALUE = SynthData::LFO_TIMER_INIT_VALUE;
		static constexpr int32 _LFO_WAVE_TABLE_MASK  = SynthData::LFO_WAVE_TABLE_SIZE - 1;
		int32* _LFO_WAVE_TABLE = nullptr;
		// Noise Generator (15bit LFSR
		static constexpr int32 _LFSR_TAP_15 = 0x6000;
		int32 _lfsrReg = 1;
		int32 _noise   = 0;
	// Synthesizer Data
	public:
		VoicePart* VOICE_PARTS = nullptr;
		int32* WAVE_TABLES = nullptr;
		TonePresetWT* TONE_PRESETS = nullptr;
		SampleData* SAMPLE_DATAS = nullptr;
		SamplerPreset* SAMPLER_PRESETS = nullptr;
	
		int32 pgPhaseMask      = 0;
		int32 pgPhaseMaskNoise = 0;
		int32 pgSamplePow2tIdx = 0;  // PG PhaseStep (Pow2Table Index Offset) - SamplingRate
		int32 pgKeyPow2tIdx    = 0;  // PG PhaseStep (Pow2Table Index Offset) - ReferenceKey (NoteNumber 0)
		int32 egPow2tIdx       = 0;  // EG PhaseStep (Pow2Table Index Offset)
		int32 egPhaseStepSndOff= 0;
		int32 lfoTimerCoef     = 0;  // Coefficient Value - LFO Phase Update Timer
	// VoiceUnit Status
		// Pitch
	private:	
		int32 _ptBend     = 0;   // pitchBend
		int32 _ptBendSens = 0;   // pitchBend Sensitivity
		// LFO
		int32 _lfoPhase     = 0;  // LFO Phase
		int32 _lfoTimer     = 0;  // LFO Phase Update Timer
		int32 _lfoTimerStep = 0;  // Update Timer Step
		int32 _lfoMod       = 0;  // Modulation Depth
		int32 _lfoModSens   = 0;  // Modulation Sensitivity
		int32 _lfoOutPM     = 0;  // output - phase modulation
		int32 _lfoOutAM     = 0;  // output - amplitude modulation
		// Wave
		int32* _waveTable = nullptr;
		// EG
		/*
		- Envelop Generator State -
		[ 5674 3210 ] 8Bits Flag Status (0 = deActive)
		[ ---- RSDA ] "-" = Reserved Bits
			A = Attack  stage
			D = Decay   stage
			S = Sustain stage
			R = Release stage
		*/
		int32 _egState      = 0;
		int32* _egLevelTable= nullptr;
		int32 _egCount      = 0;
		int32 _egCountStep  = 0;
		//int32 _egCountStepAR= 0;
		int32 _egCountStepDR= 0;
		int32 _egCountStepRR= 0;
		int32 _egCntEnvSus  = 0;
		int32 _egCntEnvBtm  = 0;
		int32 _egGate       = 0;
		int32 _egSL         = 0;
		int32 _egTL         = 0;
		int32 _egTLv        = 0;
		// PG
		int32 _pgKey        = 0;
		int32 _pgBend       = 0;
		int32 _pgPhase      = 0;
		int32 _pgPhaseStep  = 0;
	// Sampler
		int32 _pgPhaseAFP = 0;  // PG Phase (After the Fixed Point)
		int32 _plbkE      = 0;  // PlayBack End Frame   [plbkE <= channelPerSamples]
		int32 _loopE      = 0;  // Loop End Frame       [loopE <= plbkE]
		int32 _loopF      = 0;  // Loop Frames          [val = loopE - lpStart]
		int32 _loopFk     = 0;  // Loop Frames          * after -> keyOff
	
	public:
		void keyOff() override;
		void soundOff() override;
	private:
		int32 _shiftEGState();
	public:
		void keyOn(const int32&, const int32&, const int32&) override;
		void keyOnSampler(const int32&, const int32&, const int32&);
		void updPartBendSens(const int32&) override;
		void updPartPitchBend(const int32&) override;
		void updPartModulation(const int32&) override;
		void updateLfoFrequency(const int32&);
	public:
		void setup(SynthData*);
	// process ---
	private:
		void _proc0(float* const&, float* const&, const int32&) override;
		void _proc1(float* const&, float* const&, const int32&) override;
		void _proc2(float* const&, float* const&, const int32&) override;
		void _proc3(float* const&, float* const&, const int32&) override;
		void _proc4(float* const&, float* const&, const int32&) override;
		void _proc5(float* const&, float* const&, const int32&) override;
	};

// Factory -----
private:
	void _resetIndexTable();
private:	
	void _recreateVU(int32 maxPoly) override;
	void _resetKeyParams() override;
// Tables -----
private:
	int32* _indexTable = nullptr;
// Synthesizer - Data -----
private:
	SynthData* _prvSynthData = nullptr;
	SynthData* _synthData = nullptr;
	int32* waveTables = nullptr;  // 128 * 128
	SampleData* sampleDatas = nullptr;
public:
	TonePresetWT* tonePresets = nullptr;  // [0 - 127](ProgramChange), [128 - 255](Percussion Key)
	SamplerPreset* samplerPresets = nullptr;
private:
	int32 _waveTableSizeBits = 5;
	int32 _waveTableSize     = 1 << _waveTableSizeBits;
	int32 _maxWaveValueBits  = 8;
public:
	int32 getWaveTableSize() { return _waveTableSize; }
	int32 setWaveTableSize(int32 size);
	void getWaveForm(int32 id, uint8* buffer, int32 bufferSize, int32 bitDepth);
	void getWaveForm(int32 id, int32* buffer, int32 bufferSize, int32 bitDepth);
	void getWaveFormN(int32 id, float* buffer, int32 bufferSize);
	int32 getWaveTableValue(int32 id, int32 index, int32 bitDepth);
	float getWaveTableValueN(int32 id, int32 index);
	// wave [0-255(128 = 0)](bitDepth = 8)
	void setWaveForm(int32 id, const uint8* buffer, int32 bufferSize, int32 bitDepth);
	void setWaveForm(int32 id, const int32* buffer, int32 bufferSize, int32 bitDepth);
	// wave [-1.0 ~ 1.0]
	void setWaveFormN(int32 id, const float* buffer, int32 bufferSize, int32 bitDepth = 8);
	// id[0-127], index[0-127], value[0-255(by bitDepth -> 8)], bitDepth[1-8]
	void setWaveTableValue(int32 id, int32 index, uint8 value, int32 bitDepth);
	// id[0-127], index[0-127], value[0-255(by bitDepth -> 8)], bitDepth[1-8]
	void setWaveTableValue(int32 id, int32 index, int32 value, int32 bitDepth);
	// id[0-127], index[0-127], value[-1.0~1.0]
	void setWaveTableValueN(int32 id, int32 index, float value, int32 bitDepth = 8);
	SampleData* getSampleData(int32 id){ return &sampleDatas[id]; }
	void setSampleData(int32 id, SampleData&& sampleData);
	void setSampleData(int32 id, const SampleData& sampleData, bool deepCopy = true);
	void loadPCMSamples(int32 id, float* wave, int32 samples, int32 ch);
	void loadPCMData(int32 id, uint8* dat, int32 sampleFrames, int32 ch, int32 bitRate, bool sign, bool bigEndian);
	void clearSampleData(int32 id);
	void factoryDataReset() override;
// Synthesizer - main -----
private:
	VoiceUnitWT* _voiceUnitExt = nullptr;
// Constructor
public:
	SynthesizerWT() : SynthesizerWT(32, 32, 440.0f, 1.0f, true){}
	SynthesizerWT(int32 tblSize = 32, int32 maxPoly = 32, float refPitch = 440.0f, float masVol = 1.0f, bool privateTableData = true);// Controls ---
// Copy Constructor
	SynthesizerWT(const SynthesizerWT&) = delete;
	SynthesizerWT& operator = (const SynthesizerWT&) = delete;
// Move Constructor
	SynthesizerWT(SynthesizerWT&&)noexcept;
	SynthesizerWT& operator = (SynthesizerWT&&)noexcept;
// Destructor
	~SynthesizerWT();
// MidiEvent ---
private:	
	void _noteOn(const MidiEvent&) override;
	void _readSysExData(const MidiEvent&) override;
};

}}}
#endif